@@ -52,7 +52,10 @@ following example plugin:
 import enum
 import keyword
 import sys
-from typing import BinaryIO, Callable, Dict, List, Optional, Set
+from typing import BinaryIO, Callable, Dict, List, Optional, Set, Tuple
+from operator import ior
+from functools import reduce
+
 import google.protobuf.descriptor_pool
 import google.protobuf.descriptor_pb2
 import google.protobuf.message_factory
@@ -260,31 +263,106 @@ class Registry:
                 enums.append(enum)
         return enums
 
+    def resolve_enum_type(
+        self, reference_scope: str, proto_name: str
+    ) -> Optional["Enum"]:
+        """Resolve an enum name to an enum.
 
-def _resolve_enum_type_name(
-    registry: Registry, ref: str, type_name: str
-) -> Optional["Enum"]:
-    """Find a enum in the registry.
+        Searches for an enum within the registry by its proto name. If the
+        `proto_name` has a leading dot the name is treated as fully qualified,
+        otherwise the enum is resolved relative to the reference scope using
+        C++ scoping rules.
 
-    Intended to be used to resolve `FieldDescriptorProto.type_name`.
-    """
-    if type_name.startswith("."):
-        return registry.enum_by_name(type_name[1:])
-    return Exception("resolve without dot notation not implemented yet")
+        E.g. given a `reference_scope` of `"mycom.cloud.datastore.v1.Hello`" and
+        a `proto_name' of `"World"` the registry would be search for (in that
+        order):
 
+            - mycom.cloud.datastore.v1.Hello.World
+            - mycom.cloud.datastore.v1.World
+            - mycom.cloud.datastore.World
+            - mycom.cloud.World
+            - mycom.World
+            - World
 
-def _resolve_message_type_name(
-    registry: Registry, ref: str, type_name: str
-) -> Optional["Message"]:
-    """Resolve a type_name.
+        and the first existing enum type would be returned.
 
-    Intended to be used to resolve `FieldDescriptorProto.type_name`.
-    `MethodDescriptorProto.input_type` and `MethodDescriptorProto-output_type`.
-    To resolve this, several
-    """
-    if type_name.startswith("."):
-        return registry.message_by_name(type_name[1:])
-    return None
+        Arguments
+        ---------
+        reference_scope : str
+            The current scope that acts as starting points in the enum type
+            resolution process.
+        proto_name : str
+            The proto (enum type) name to resolve.
+
+        Returns
+        -------
+        response: protogen.Enum | None
+            The resolved protogen enum type, or `None` if no enum with that name
+            could be found under the reference scope.
+        """
+        if proto_name.startswith("."):
+            return self.enum_by_name(proto_name[1:])
+
+        ref_split = reference_scope.split(".")
+        for i in reversed(range(len(ref_split) + 1)):
+            ref_i = ".".join(ref_split[:i]) + "." + proto_name
+            enum = self.enum_by_name(ref_i)
+            if enum is not None:
+                return enum
+
+        return None
+
+    def resolve_message_type(
+        self, reference_scope: str, proto_name: str
+    ) -> Optional["Message"]:
+        """Resolve a message name to a message.
+
+        Searches for a message within the registry by its proto name. If the
+        `proto_name` has a leading dot the name is treated as fully qualified,
+        otherwise the message is resolved relative to the reference scope using
+        C++ scoping rules.
+
+        E.g. given a `reference_scope` of `"mycom.cloud.datastore.v1.Hello`" and
+        a `proto_name' of `"World"` the registry would be search for (in that
+        order):
+
+            - mycom.cloud.datastore.v1.Hello.World
+            - mycom.cloud.datastore.v1.World
+            - mycom.cloud.datastore.World
+            - mycom.cloud.World
+            - mycom.World
+            - World
+
+        and the first existing message type would be returned.
+
+        Arguments
+        ---------
+        reference_scope : str
+            The current scope that acts as starting points in the message type
+            resolution process.
+        proto_name : str
+            The proto (message type) name to resolve.
+
+        Returns
+        -------
+        response: protogen.Message | None
+            The resolved protogen message type, or `None` if no message with
+            that name could be found under the reference scope.
+        """
+        if proto_name.startswith("."):
+            return self.message_by_name(proto_name[1:])
+
+        ref_split = reference_scope.split(".")
+        for i in reversed(range(len(ref_split) + 1)):
+            prefix = ref_split[:i]
+            _name = prefix + [proto_name]
+            name = ".".join(_name)
+
+            msg = self.message_by_name(name)
+            if msg is not None:
+                return msg
+
+        return None
 
 
 def _clean_comment(cmmt: str) -> str:
@@ -780,8 +858,8 @@ class Field:
     def _resolve(self, registry: Registry):
         # resolve extendee is present
         if self.proto.HasField("extendee"):
-            self.extendee = _resolve_message_type_name(
-                registry, self.full_name, self.proto.extendee
+            self.extendee = registry.resolve_message_type(
+                self.full_name, self.proto.extendee
             )
             if self.extendee is None:
                 raise ResolutionError(
@@ -797,9 +875,7 @@ class Field:
                     full_name=self.full_name,
                     msg="is of kind ENUM but has no `type_name` set",
                 )
-            self.enum = _resolve_enum_type_name(
-                registry, self.full_name, self.proto.type_name
-            )
+            self.enum = registry.resolve_enum_type(self.full_name, self.proto.type_name)
             if not self.enum:
                 raise ResolutionError(
                     file=self.parent_file.proto.name,
@@ -814,8 +890,8 @@ class Field:
                     full_name=self.full_name,
                     msg="is of kind MESSAGE but has no `type_name` set",
                 )
-            self.message = _resolve_message_type_name(
-                registry, self.full_name, self.proto.type_name
+            self.message = registry.resolve_message_type(
+                self.full_name, self.proto.type_name
             )
             if self.message is None:
                 raise ResolutionError(
@@ -912,13 +988,14 @@ class Message:
         path: List[int],
     ):
         self.proto = proto
-        self.py_ident = parent_file.py_import_path.ident(
-            self.proto.name
-        )  # TODO how to handle nested messages?
-        if parent is not None:
-            self.full_name = parent.full_name + "." + proto.name
-        else:
+        if parent is None:
             self.full_name = parent_file.proto.package + "." + proto.name
+            self.py_ident = parent_file.py_import_path.ident(self.proto.name)
+        else:
+            self.full_name = parent.full_name + "." + proto.name
+            self.py_ident = parent_file.py_import_path.ident(
+                parent.py_ident.py_name + "." + proto.name
+            )
         self.parent_file = parent_file
         self.parent = parent
         self.location = _resolve_location(parent_file.proto, path)
@@ -1093,8 +1170,8 @@ class Method:
         self.location = _resolve_location(parent.parent_file.proto, path)
 
     def _resolve(self, registry: Registry):
-        self.input = _resolve_message_type_name(
-            registry, self.full_name, self.proto.input_type
+        self.input = registry.resolve_message_type(
+            self.full_name, self.proto.input_type
         )
         if self.input is None:
             raise ResolutionError(
@@ -1102,8 +1179,8 @@ class Method:
                 desc=self.full_name,
                 ref=self.proto.input_type,
             )
-        self.output = _resolve_message_type_name(
-            registry, self.full_name, self.proto.output_type
+        self.output = registry.resolve_message_type(
+            self.full_name, self.proto.output_type
         )
         if self.output is None:
             raise ResolutionError(
@@ -1273,6 +1350,59 @@ def _indent(s: str, width: int) -> str:
     return "\n".join(prefix_lines)
 
 
+class CodeGeneratorResponse:
+    """A code generator response.
+
+    This is the ``protogen`` equivalent to a protobuf CodeGeneratorResponse.
+
+    Attributes
+    ----------
+    proto : google.protobuf.descriptor_pb2.CodeGeneratorResponse
+        The raw CodeGeneratorResponse.
+    """
+
+    def __init__(
+        self, proto: google.protobuf.compiler.plugin_pb2.CodeGeneratorResponse
+    ) -> None:
+        self.proto = proto
+
+    def has_file(self, filename: str) -> bool:
+        """Checks if a file in the CodeGeneratorResponse.
+
+        Arguments
+        ---------
+        filename : str
+            Name of the file to check.
+
+        Returns
+        -------
+        ok: bool
+            `True`, if the file is contained in the response, `False` otherwise.
+        """
+        return any([file.name == filename for file in self.proto.file])
+
+    def file_content(self, filename) -> Tuple[str, bool]:
+        """Returns the content of a file from the CodeGeneratorResponse.
+
+        Arguments
+        ---------
+        filename : str
+            Name of the file to get the content for.
+
+        Returns
+        -------
+        content: Tuple[bool, str]
+            Returns `True` and the content of the file if a file with that name
+            exists in the CodeGeneratorResponse. Otherwise `False` and the empty
+            string is returned.
+        """
+        for file in self.proto.file:
+            if file.name == filename:
+                return file.content, True
+
+        return "", False
+
+
 class GeneratedFile:
     """An output buffer to write generated code to.
 
@@ -1574,8 +1704,11 @@ class Options:
         self,
         *,
         py_import_func: Callable[[str, str], PyImportPath] = default_py_import_func,
-        input: BinaryIO = sys.stdin.buffer,
-        output: BinaryIO = sys.stdout.buffer,
+        input: Optional[BinaryIO] = None,
+        output: Optional[BinaryIO] = None,
+        supported_features: List[
+            "google.protobuf.compiler.plugin_pb2.CodeGeneratorResponse.Feature"
+        ] = [],
     ):
         """Create options for the resolution process.
 
@@ -1590,14 +1723,26 @@ class Options:
             to use :func:`default_py_import_func`.
         input : BinaryIO, optional
             The input stream to read the CodeGeneratorRequest from. Defaults
-            to :attr:`sys.stdin.buffer`.
+            to :attr:`sys.stdin.buffer` if set as None.
         output : BinaryIO, optional
             The output stream to write the CodeGeneratorResponse to.
-            Defaults to :attr:`sys.stdout.buffer`.
+            Defaults to :attr:`sys.stdout.buffer` if set as None.
+        supported_features : List[str]
+            List of features that are supported by the plugin. This list will be
+            delegated to protoc via the CodeGeneratorresponse.supported_features
+            field.
+            For example, to indicate that the plugin supports optionals, provide
+            `google.protobuf.compiler.plugin_pb2.CodeGeneratorResponse.Feature.FEATURE_PROTO3_OPTIONAL`.
+            in the list.
         """
         self._input = input
+        if input is None:
+            self._input = sys.stdin.buffer
         self._output = output
+        if output is None:
+            self._output = sys.stdout.buffer
         self._py_import_func = py_import_func
+        self._supported_features = supported_features
 
     def run(self, f: Callable[[Plugin], None]):
         """Start resolution process and run ``f`` with the :class:`Plugin` containing the resolved classes.
@@ -1674,4 +1819,7 @@ class Options:
 
         # Write response.
         resp = plugin._response()
+        if len(self._supported_features) > 0:
+            resp.supported_features = reduce(ior, self._supported_features)
         self._output.write(resp.SerializeToString())
+        self._output.flush()
